# coding=UTF-8
"""
Return Python code used to serialize/unserialize/get_size of lists of fields.

If you are going to paste the code you get here, note you nede to paste it into
a module that has following ok:

    * local variables denoted with a list of Field (namedtuple) with optional prefix exist
    * function header was already emitted
    * indent_level is in multiple of fours
    * following imports exists:
        * import struct
        * from coolamqp.framing.field_table import enframe_table, deframe_table, frame_table_size
    * local variables buf and offset exist
    * local variable start_offset can be created

"""
from __future__ import absolute_import, division, print_function

import math

from coolamqp.framing.base import BASIC_TYPES
from coolamqp.framing.compilation.utilities import format_field_name


def get_counter(fields, prefix=u'', indent_level=2):
    """
    Emit code that counts how long this struct is.

    :param fields: list of Field instances
    :param prefix: pass "self." is inside a class
    :param indent_level: amount of tabs
    :return: block of code that does that
    """

    parts = []
    accumulator = 0
    bits = 0
    for field in fields:
        bt = field.basic_type
        nam = prefix + format_field_name(field.name)

        if (bits > 0) and (bt != 'bit'):  # sync bits if not
            accumulator += int(math.ceil(bits / 8))
            bits = 0

        if field.basic_type == 'bit':
            bits += 1
        elif field.reserved:
            accumulator += BASIC_TYPES[field.basic_type][3]
        elif BASIC_TYPES[bt][0] is not None:
            accumulator += BASIC_TYPES[field.basic_type][0]
        elif bt == 'shortstr':
            parts.append('len(' + nam + ')')
            accumulator += 1
        elif bt == 'longstr':
            parts.append('len(' + nam + ')')
            accumulator += 4
        elif bt == 'table':
            parts.append('frame_table_size(' + nam + ')')
            accumulator += 0  # because frame_table_size accounts for that 4 leading bytes
        else:
            raise Exception()

    if bits > 0:  # sync bits
        accumulator += int(math.ceil(bits / 8))

    return (u'    ' * indent_level) + u'return ' + (
        u' + '.join([str(accumulator)] + parts)) + u'\n'


 # type: (...) -> tp.Tuple[str, dict]
def get_from_buffer(fields, prefix='', indent_level=2, remark=False):
    """
    Emit code that collects values from buf:offset, updating offset as progressing.
    :param remark: BE FUCKING VERBOSE! #DEBUG
    """
    code = []
    structers = {}

    def emit(fmt, *args):
        args = list(args)
        code.append(u'    ' * indent_level)
        assert fmt.count('%s') == len(args)
        for arg in args:
            fmt = fmt.replace('%s', str(arg), 1)
        code.append(fmt)
        code.append('\n')

    # actually go and load it

    bits = []
    ln = {'ln': 0}  # so I can modify from outside
    to_struct = []

    def emit_bits():
        if len(bits) == 0:
            return
        if remark:
            print('Bits are being banged')
        if all(n == '_' for n in bits):
            # everything is reserved, lol
            emit('offset += 1')
        else:
            to_struct.append(('_bit', 'B'))
            emit_structures(dont_do_bits=True)

            for multiplier, bit in enumerate(bits):
                if bit != '_':
                    emit("%s = bool(_bit >> %s)", bit, multiplier)
            emit('offset += 1')

        del bits[:]

    def emit_structures(dont_do_bits=False): # type: (bool) -> dict
        if not dont_do_bits:
            emit_bits()
        if len(to_struct) == 0:
            return {}
        fffnames = [a for a, b in to_struct if a != u'_']  # skip reserved
        ffffmts = [b for a, b in to_struct]
        fmts = u''.join(ffffmts)
        emit("%s, = STRUCT_%s.unpack_from(buf, offset)",
             u', '.join(fffnames), fmts)
        emit("offset += %s", ln['ln'])
        ln['ln'] = 0
        del to_struct[:]
        return {fmts: fmts}

    for field in fields:
        fieldname = prefix + format_field_name(field.name)

        if (len(bits) > 0) and (field.basic_type != u'bit'):
            emit_bits()

        if remark:
            print('Doing', fieldname, 'of type', field.basic_type)

        # offset is current start
        # length is length to read
        if BASIC_TYPES[field.basic_type][0] is not None:
            # static type shit has

            assert len(bits) == 0

            if field.reserved:
                to_struct.append(
                    (u'_', '%sx' % (BASIC_TYPES[field.basic_type][0],)))
            else:
                to_struct.append((fieldname, BASIC_TYPES[field.basic_type][1]))

            ln['ln'] += BASIC_TYPES[field.basic_type][0]
        elif field.basic_type == u'bit':
            bits.append('_' if field.reserved else fieldname)
        elif field.basic_type == u'table':  # oh my god
            structers.update(emit_structures())

            assert len(bits) == 0
            assert len(to_struct) == 0

            emit("%s, delta = deframe_table(buf, offset)", fieldname)
            emit("offset += delta")
        else:  # longstr or shortstr
            f_q, f_l = ('L', 4) if field.basic_type == u'longstr' else ('B', 1)
            to_struct.append(('s_len', f_q))
            ln['ln'] += f_l
            structers.update(emit_structures())
            if field.reserved:
                emit("offset += s_len # reserved field!")
            else:
                emit("%s = buf[offset:offset+s_len]", fieldname)
                emit("offset += s_len")

        # check bits for overflow
        if len(bits) == 8:
            emit_bits()

    structers.update(emit_structures())

    return u''.join(code), structers


def get_serializer(fields, prefix='', indent_level=2):      # type: (list, str) -> str, dict
    """
    Emit code that serializes the fields into buf at offset

    :param fields: list of Field instances
    :param prefix: pass "self." is inside a class
    :return: block of code that does that, dictionary of struct-ers
    """
    code = []
    structers = {}

    def emit(fmt, *args):
        args = list(args)
        code.append(u'    ' * indent_level)
        while len(args) > 0:
            fmt = fmt.replace('%s', args[0], 1)
            del args[0]
        code.append(fmt)
        code.append('\n')

    formats = []
    format_args = []
    bits = []

    def emit_bits():
        p = []
        formats.append('B')
        if all(bit_name == 'False' for bit_name in bits):
            format_args.append('0')
        else:
            for bit_name, modif in zip(bits, range(8)):
                if bit_name != 'False':
                    p.append('(' + bit_name + ' << %s)' % (
                        modif,))  # yes you can << bools
            format_args.append(u' | '.join(p))
        del bits[:]

    def emit_single_struct_pack():
        formats_str = u''.join(formats)
        if formats_str not in structers:
            structers[formats_str] = formats_str
        emit("buf.write(STRUCT_%s.pack(%s))", formats_str,
             u', '.join(format_args))
        del formats[:]
        del format_args[:]

    for field in fields:
        nam = prefix + format_field_name(field.name)

        if (len(bits) == 8) or ((len(bits) > 0) and field.basic_type != 'bit'):
            emit_bits()

        if field.basic_type == 'bit':
            if field.reserved:
                bits.append("False")
            else:
                bits.append(nam)
        elif field.reserved:
            # Just pasta
            emit('buf.write(%s)', BASIC_TYPES[field.basic_type][2])
        else:
            if field.basic_type in ('shortstr', 'longstr'):
                formats.append('B' if field.basic_type == 'shortstr' else 'I')
                format_args.append('len(' + nam + ')')
                emit_single_struct_pack()
                emit('buf.write(%s)', nam)
            elif field.basic_type == 'table':
                if len(bits) > 0:
                    emit_bits()
                if len(formats) > 0:
                    emit_single_struct_pack()
                emit('enframe_table(buf, %s)', nam)
            else:
                formats.append(BASIC_TYPES[field.basic_type][1])
                format_args.append(nam)

    if len(bits) > 0:
        emit_bits()
    if len(formats) > 0:
        emit_single_struct_pack()

    emit('')  # eol

    return u''.join(code), structers



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import functools
import logging
import threading

logger = logging.getLogger(__name__)


class ConfirmableRejectable(object):
    """
    Protocol for objects put into AtomicTagger. You need not subclass it,
    just support this protocol.
    """
    __slots__ = ()

    def confirm(self):  # type: () -> None
        """
        This has been ACK'd
        :return: don't care
        """

    def reject(self):  # type: () -> None
        """
        This has been REJECT'd/NACK'd
        :return: don't care
        """


class FutureConfirmableRejectable(ConfirmableRejectable):
    """
    A ConfirmableRejectable that can result a future (with None),
    or Exception it with a message
    """
    __slots__ = ('future', )

    def __init__(self, future):  # type: (concurrent.futures.Future) -> None
        self.future = future

    def confirm(self):  # type: () -> None
        self.future.set_result(None)

    def reject(self):  # type: () -> None
        self.future.set_exception(Exception())


class AtomicTagger(object):
    """
    This implements a thread-safe dictionary of (integer=>ConfirmableRejectable | None),
    used for processing delivery tags / (negative) acknowledgements.
        - you can requisition a key. This key belongs only to you, and the whole world
          doesn't know you have it.

            delivery_tag_to_use = tagger.get_key()

        - you can deposit a ConfirmableRejectable into the  tagger.

            tagger.deposit(delivery_tag, message)

         After you do so, this tag is subject to be acked/nacked. Read on.

        - you can (multiple)(ack/nack) messages. This coresponds to multiple bit
          used in basic.ack/basic.nack.

          If this is done, your message objects (that MUST implement the
          ConfirmableRejectable protocol) will have respective methods called.
          These methods MUST NOT depend on particular state of locking by this
          object.

    Thread safety is implemented using reentrant locking. The lock object is a
    threading.RLock, and you can access it at atomicTagger.lock.

    Please note that delivery tags are increasing non-negative integer.
    Therefore, X>Y implies that sending/receiving X happened after Y.

    Note that key/delivery_tag of 0 has special meaning of "everything so far".

    This has to be fast for most common cases. Corner cases will be resolved correctly,
    but maybe not fast.
    """
    __slots__ = ('lock', 'next_tag', 'tags')

    def __init__(self):
        self.lock = threading.RLock()

        # Protected by lock
        self.next_tag = 1  # 0 is AMQP-reserved to mean "everything so far"
        self.tags = []  # a list of (tag, ConfirmableRejectable)
        # they remain to be acked/nacked
        # invariant: FOR EACH i, j: (i>j) => (tags[i][0] > tags[j][0])

    def deposit(self, tag, obj):
        """
        Put a tag into the tag list.

        Putting the same tag more than one time will result in undefined behaviour.

        :param tag: non-negative integer
        :param obj: ConfirmableRejectable
                    if you put something that isn't a ConfirmableRejectable, you won't get bitten
                    until you call .ack() or .nack().
        """
        assert tag >= 0
        opt = (tag, obj)

        with self.lock:
            if len(self.tags) == 0:
                self.tags.append(opt)
            elif self.tags[-1][0] < tag:
                self.tags.append(opt)
            else:
                # Insert a value at place where it makes sense. Iterate from the end, because
                # values will usually land there...
                i = len(self.tags) - 1  # start index

                while i > 0:  # this will terminate at i=0
                    if self.tags[i][
                        0] > tag:  # this means we should insert it here...
                        break
                    i -= 1  # previousl index

                self.tags.insert(i, opt)

    def __acknack(self, tag, multiple, ack):
        """
        :param tag: Note that 0 means "everything"
        :param ack: True to ack, False to nack
        """
        # Compute limits - they go from 0 to somewhere
        with self.lock:
            start = 0
            # start and stop will signify the PYTHON SLICE parameters

            if tag > 0:

                if multiple:
                    # Compute the ranges
                    for stop, opt in enumerate(self.tags):
                        if opt[0] == tag:
                            stop += 1  # this is exactly this tag. Adjust stop to end one further (Python slicing) and stop
                            break
                        if opt[0] > tag:
                            break  # We went too far, but it's OK, we don't need to bother with adjusting stop
                    else:
                        # List finished without breaking? That would mean the entire range!
                        stop = len(self.tags)
                else:
                    # Just find that piece
                    for index, opt in enumerate(self.tags):
                        if opt[0] == tag:
                            stop = index + 1
                            break
                    else:
                        return  # not found!

                if not multiple:
                    start = stop - 1
            else:
                # Oh, I know the range!
                stop = len(self.tags)

            items = self.tags[start:stop]
            del self.tags[start:stop]

        for tag, cr in items:
            if ack:
                cr.confirm()
            else:
                cr.reject()

    def ack(self, tag, multiple):
        """
        Acknowledge given objects.

        If multiple, objects UP TO AND INCLUDING tag will have .confirm() called.
        If it's false, only this precise objects will have done so.
        It this object does not exist, nothing will happen. Acking same tag more than one time
        is a no-op.

        Things acked/nacked will be evicted from .data
        :param tag: delivery tag to use. Note that 0 means "everything so far"
        """
        self.__acknack(tag, multiple, True)

    def nack(self, tag, multiple):
        """
        Acknowledge given objects.

        If multiple, objects UP TO AND INCLUDING tag will have .confirm() called.
        If it's false, only this precise objects will have done so.
        It this object does not exist, nothing will happen. Acking same tag more than one time
        is a no-op.

        Things acked/nacked will be evicted from .data
        :param tag: delivery tag to use. Note that 0 means "everything so far"
        """
        self.__acknack(tag, multiple, False)

    def get_key(self):
        """
        Return a key. It won't be seen here until you deposit it.

        It's just yours, and you can do whatever you want with it, even drop on the floor.
        :return: a positive integer
        """
        with self.lock:
            self.next_tag += 1
            return self.next_tag - 1


class Synchronized(object):
    """
    I have a lock and can sync on it. Use like:

    class Synced(Synchronized):

        @synchronized
        def mandatorily_a_instance_method(self, ...):
            ...

    """

    def __init__(self):
        self._monitor_lock = threading.Lock()

    def get_monitor_lock(self):
        return self._monitor_lock

    @staticmethod
    def synchronized(fun):
        @functools.wraps(fun)
        def monitored(*args, **kwargs):
            with args[0]._monitor_lock:
                return fun(*args, **kwargs)

        return monitored



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import collections
import logging
import socket
import time
import uuid
import typing as tp
import monotonic

from coolamqp.exceptions import ConnectionDead
from coolamqp.framing.base import AMQPMethodPayload
from coolamqp.framing.definitions import ConnectionClose, ConnectionCloseOk
from coolamqp.framing.frames import AMQPMethodFrame
from coolamqp.objects import Callable
from coolamqp.uplink.connection.recv_framer import ReceivingFramer
from coolamqp.uplink.connection.send_framer import SendingFramer
from coolamqp.uplink.connection.states import ST_ONLINE, ST_OFFLINE, \
    ST_CONNECTING
from coolamqp.uplink.connection.watches import MethodWatch
from coolamqp.uplink.handshake import Handshaker

logger = logging.getLogger(__name__)


def alert_watches(watches, trigger):
    """
    Notify all watches in this collection.

    Return a list of alive watches.
    :param watches: list of Watch
    :return: tuple of (list of Watch, bool - was any watch fired?)
    """
    watch_handled = False
    alive_watches = []
    while len(watches) > 0:
        watch = watches.pop()

        if watch.cancelled:
            continue

        watch_triggered = watch.is_triggered_by(trigger)
        watch_handled |= watch_triggered

        if watch.cancelled:
            continue

        if not any((watch_triggered, watch.oneshot, watch.cancelled)):
            # Watch remains alive if it was NOT triggered, or it's NOT a oneshot or it's not cancelled
            alive_watches.append(watch)
        elif not watch.oneshot and not watch.cancelled:
            alive_watches.append(watch)
        elif watch.oneshot and not watch_triggered:
            alive_watches.append(watch)

    if set(alive_watches) != set(watches):
        for removed_watch in set(watches)-set(alive_watches):
            logger.debug('Removing watch %s', repr(removed_watch))
    return alive_watches, watch_handled


class Connection(object):
    """
    An object that manages a connection in a comprehensive way.

    It allows for sending and registering watches for particular things. Watch will
    listen for eg. frame on particular channel, frame on any channel, or connection teardown.
    Watches will also get a callback for connection being non-operational (eg. torn down).

    WARNING: Thread-safety of watch operation hinges on atomicity
    of .append and .pop.

    Lifecycle of connection is such:

        Connection created  ->  state is ST_CONNECTING
        .start() called     ->  state is ST_CONNECTING
        connection.open-ok  ->  state is ST_ONLINE

    This logger is talkative mostly on INFO, and regarding connection state
    """

    def __init__(self, node_definition,  # type: coolamqp.objects.NodeDefinition
                 listener_thread, extra_properties,  # type: tp.Dict[bytes, tp.Tuple[tp.Any, str]]
                 log_frames=None,
                 name=None          
                 ):
        """
        Create an object that links to an AMQP broker.

        No data will be physically sent until you hit .start()

        :param node_definition: NodeDefinition instance to use
        :param listener_thread: ListenerThread to use as async engine
        :type listener_thread: coolamqp.uplink.listener.ListenerThread
        :param extra_properties: extra properties to send to the target server
            must conform to the syntax given in (/coolamqp/uplink/handshake.py)'s CLIENT_PROPERTIES
        """
        self.listener_thread = listener_thread
        self.node_definition = node_definition
        self.uuid = uuid.uuid4().hex[:5]
        self.name = name or 'CoolAMQP'
        self.recvf = ReceivingFramer(self.on_frame)
        self.extra_properties = extra_properties
        # todo a list doesn't seem like a very strong atomicity guarantee
        self.watches = {}  # channel => list of [Watch instance]
        self.any_watches = []  # list of Watches that should check everything

        self.finalize = Callable(oneshots=True)  #: public

        self.state = ST_CONNECTING

        self.callables_on_connected = []  # list of callable/0

        # Negotiated connection parameters - handshake will fill this in
        self.free_channels = []  # attaches can use this for shit.
        # WARNING: thread safety of this hinges on atomicity of .pop or .append
        self.frame_max = None
        self.heartbeat = None
        self.extensions = []

        # To be filled in later
        self.listener_socket = None
        self.sendf = None

        # To log frames
        self.log_frames = log_frames

    def call_on_connected(self, callable):
        """
        Register a callable to be called when this links to the server.

        If you call it while the connection IS up, callable will be called even before this returns.

        You should be optimally an attached attache to receive this.

        :param callable: callable/0 to call
        """
        if self.state == ST_ONLINE:
            callable()
        else:
            self.callables_on_connected.append(callable)

    def on_connected(self):
        """Called by handshaker upon reception of final connection.open-ok"""
        logger.info('[%s] Connection ready.', self.name)

        self.state = ST_ONLINE

        while len(self.callables_on_connected) > 0:
            self.callables_on_connected.pop()()

    def start(self, timeout):
        """
        Start processing events for this connect. Create the socket,
        transmit 'AMQP\x00\x00\x09\x01' and roll.

        Warning: This will block for as long as the TCP connection setup takes.
        """

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        start_at = monotonic.monotonic()
        while True:
            try:
                sock.connect(
                    (self.node_definition.host, self.node_definition.port))
            except socket.error as e:
                time.sleep(0.5)  # Connection refused? Very bad things?
                if monotonic.monotonic() - start_at > timeout:
                    raise ConnectionDead()
            else:
                break

        logger.debug('[%s] TCP connection established, authentication in progress', self.name)

        sock.settimeout(0)
        header = bytearray(b'AMQP\x00\x00\x09\x01')
        rest = sock.send(header)
        while rest < len(header):
            time.sleep(0.1)
            header = header[rest:]
            rest = sock.send(header)

        self.watch_for_method(0, (ConnectionClose, ConnectionCloseOk),
                              self.on_connection_close)

        # Note that these are placed in just the right order. Sometimes there would
        # be a race condition that ConnectionStart has arrived before there could
        # be a watch for it set
        self.listener_socket = self.listener_thread.register(sock,
                                                             on_read=self.recvf.put,
                                                             on_fail=self.on_fail)
        self.sendf = SendingFramer(self.listener_socket.send)
        Handshaker(self, self.node_definition, self.on_connected, self.extra_properties)
        self.listener_thread.activate(self.listener_socket)

    def on_fail(self):
        """
        Called by event loop when the underlying connection is closed.

        This means the connection is dead, cannot be used anymore, and all operations
        running on it now are aborted, null and void.

        This calls fails all registered watches.
        Called by ListenerThread.

        WARNING: Note that .on_fail can get called twice - once from .on_connection_close,
        and second time from ListenerThread when socket is disposed of
        Therefore we need to make sure callbacks are called EXACTLY once
        """
        logger.info('[%s] Connection lost', self.name)

        self.state = ST_OFFLINE  # Update state

        watchlists = [self.watches[channel] for channel in self.watches]

        for watchlist in watchlists:  # Run all watches - failed
            for watch in watchlist:
                if not watch.cancelled:
                    watch.failed()

        for watch in self.any_watches:
            if not watch.cancelled:
                watch.failed()

        self.watches = {}  # Clear the watch list
        self.any_watches = []

        # call finalizers
        self.finalize()

    def on_connection_close(self, payload):
        """
        Server attempted to close the connection.. or maybe we did?

        Called by ListenerThread.
        """
        self.on_fail()  # it does not make sense to prolong the agony

        if isinstance(payload, ConnectionClose):
            self.send([AMQPMethodFrame(0, ConnectionCloseOk())])
            logger.info(u'[%s] Broker closed our connection - code %s reason %s',
                        self.name,
                        payload.reply_code,
                        payload.reply_text.tobytes().decode('utf8'))

        elif isinstance(payload, ConnectionCloseOk):
            self.send(None)

    def send(self, frames, priority=False):
        """
        Schedule to send some frames.

        Take care: This won't stop you from sending frames larger tham frame_max.
        Broker will probably close the connection if he sees that.

        :param frames: list of frames or None to close the link
        :param reason: optional human-readable reason for this action
        """
        if self.log_frames is not None:
            for frame in frames:
                self.log_frames.on_frame(time.monotonic(), frame, 'to_server')

        if frames is not None:
            self.sendf.send(frames, priority=priority)
        else:
            # Listener socket will kill us when time is right
            self.listener_socket.send(None)

    def on_frame(self, frame):
        """
        Called by event loop upon receiving an AMQP frame.

        This will verify all watches on given channel if they were hit,
        and take appropriate action.

        Unhandled frames will be logged - if they were sent, they probably were important.

        :param frame: AMQPFrame that was received
        """
        if self.log_frames is not None:
            self.log_frames.on_frame(time.monotonic(), frame, 'to_client')

        watch_handled = False  # True if ANY watch handled this

        if isinstance(frame, AMQPMethodFrame):
            logger.debug('[%s] Received %s', self.uuid, frame.payload.NAME)

        # ==================== process per-channel watches
        #
        #   Note that new watches may arrive while we process existing watches.
        #   Therefore, we need to copy watches and zero the list before we proceed
        if frame.channel in self.watches:
            watches = self.watches[frame.channel]  # a list
            self.watches[frame.channel] = []

            alive_watches, f = alert_watches(watches, frame)
            watch_handled |= f

            if frame.channel in self.watches:
                # unwatch_all might have gotten called, check that
                for watch in alive_watches:
                    self.watches[frame.channel].append(watch)

        # ==================== process "any" watches
        any_watches = self.any_watches
        self.any_watches = []
        alive_watches, f = alert_watches(any_watches, frame)

        watch_handled |= f

        for watch in alive_watches:
            self.any_watches.append(watch)

        if not watch_handled:
            if isinstance(frame, AMQPMethodFrame):
                logger.warning('[%s] Unhandled method frame %s', self.name, repr(frame.payload))
            else:
                logger.warning('[%s] Unhandled frame %s', self.name, frame)

    def watchdog(self, delay, callback):
        """
        Call callback in delay seconds. One-shot.

        Shall the connection die in the meantime, watchdog will not
        be called, and everything will process according to
        ListenerThread's on_fail callback.

        This is necessary to implement timeout detection when setting up the connection
        and heartbeat is not yet configured.
        """
        try:
            self.listener_socket.oneshot(delay, callback)
        except AttributeError:
            pass  # print(dir(self))

    def unwatch_all(self, channel_id):
        """
        Remove all watches from specified channel
        """
        self.watches.pop(channel_id, None)

    def watch(self, watch):
        """
        Register a watch.
        :param watch: Watch to register
        """
        assert self.state != ST_OFFLINE
        if watch.channel is None:
            self.any_watches.append(watch)
        elif watch.channel not in self.watches:
            self.watches[watch.channel] = collections.deque([watch])
        else:
            self.watches[watch.channel].append(watch)

    def watch_for_method(self, channel, method, callback, on_fail=None):
        # type: (int, AMQPMethodPayload, tp.Callable[[AMQPMethodPayload], None],
        #   tp.Optional[tp.Callable[[AMQPMethodPayload], None]]) -> MethodWatch
        """
        :param channel: channel to monitor
        :param method: AMQPMethodPayload class or tuple of AMQPMethodPayload classes
        :param callback: callable(AMQPMethodPayload instance)
        """
        mw = MethodWatch(channel, method, callback, on_end=on_fail)
        self.watch(mw)
        return mw

    def method_and_watch(self, channel_id, method_payload, method_or_methods,
                         callback):
        """
        A syntactic sugar for

                .watch_for_method(channel_id, method_or_methdods, callback)
                .send([AMQPMethodFrame(channel_id, method_payload)])
        """
        self.watch_for_method(channel_id, method_or_methods, callback)
        self.send([AMQPMethodFrame(channel_id, method_payload)])



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import io
import logging
import typing as tp
import uuid
from concurrent.futures import Future

from coolamqp.attaches.channeler import Channeler, ST_ONLINE, ST_OFFLINE
from coolamqp.exceptions import AMQPError
from coolamqp.framing.definitions import ChannelOpenOk, BasicConsume, \
    BasicConsumeOk, QueueDeclare, QueueDeclareOk, ExchangeDeclare, \
    ExchangeDeclareOk, \
    QueueBind, QueueBindOk, ChannelClose, BasicDeliver, BasicCancel, \
    BasicAck, BasicReject, RESOURCE_LOCKED, BasicCancelOk, BasicQos, BasicQosOk
from coolamqp.framing.frames import AMQPBodyFrame, AMQPHeaderFrame
from coolamqp.objects import Callable
from coolamqp.uplink import HeaderOrBodyWatch, MethodWatch

logger = logging.getLogger(__name__)

EMPTY_MEMORYVIEW = memoryview(b'')  # for empty messages


class BodyReceiveMode(object):
    # ZC - zero copy
    # C - copy (copies every byte once)

    BYTES = 0  # message.body will be a single bytes object
    # this will gather frames as memoryviews, and b''.join() them upon
    # receiving last frame
    # this is C

    MEMORYVIEW = 1  # message.body will be returned as a memoryview object
    # this is ZC for small messages, and C for multi-frame ones
    # think less than 800B, since 2048 is the buffer for socket recv, and an
    # AMQP frame (or many frames!) have to fit there

    LIST_OF_MEMORYVIEW = 2  # message.body will be returned as list of
    # memoryview objects these constitute received pieces. this is always ZC


class Consumer(Channeler):
    """
    This object represents a consumer in the system.

    Consumer may reside on any AMQP broker, this is to be decided by CoolAMQP.
    Consumer, when created, has the state of ST_SYNCING. CoolAMQP will
    try to declare the consumer where it makes most sense for it to be.

    If it succeeds, the consumer will enter state ST_ONLINE, and callables
    on_start will be called. This means that broker has confirmed that this
    consumer is operational and receiving messages.

    Note that does not attempt to cancel consumers, or any of such nonsense.
    Having a channel per consumer gives you the unique possibility of simply
    closing  the channel. Since this implies cancelling the consumer, here you
    go.

    WARNING: READ DEFAULT VALUES IN CONSTRUCTOR! TAKE CARE WHAT YOUR CONSUMERS
     DO!

    You can subscribe to be informed when the consumer is cancelled (for any
    reason, server or client side) with:

    >>> con, fut = Cluster.consume(...)
    >>> def im_called_on_cancel_for_any_reason():   # must have arity of 0
    >>>     ..
    >>> con.on_cancel.add(im_called_on_cancel_for_any_reason)
    >>> con.cancel()

    Or, if RabbitMQ is in use, you can be informed upon a Consumer Cancel
    Notification:

    >>> con.on_broker_cancel.add(im_cancelled_by_broker)

    :param queue: Queue object, being consumed from right now.
        Note that name of anonymous queue might change at any time!
    :type queue: coolamqp.objects.Queue
    :param on_message: callable that will process incoming messages
    :type on_message: callable(ReceivedMessage instance)
    :param no_ack: Will this consumer require acknowledges from messages?
    :type no_ack: bool
    :param qos: a tuple of (prefetch size, prefetch window) for this
        consumer, or an int (prefetch window only).
        If an int is passed, prefetch size will be set to 0 (which means
        undefined), and this int will be used for prefetch window
    :type qos: tuple(int, int) or tuple(None, int) or int
    :param cancel_on_failure: Consumer will cancel itself when link goes
        down
    :type cancel_on_failure: bool
    :param future_to_notify: Future to succeed when this consumer goes
                             online for the first time.
                             This future can also raise with AMQPError if
                             it fails to.
    :type future_to_notify: concurrent.futures.Future
    :param fail_on_first_time_resource_locked: When consumer is declared
        for the first time, and RESOURCE_LOCKED is encountered, it will
        fail the future with ResourceLocked, and consumer will cancel
        itself.
        By default it will retry until success is made.
        If the consumer doesn't get the chance to be declared - because
        of a connection fail - next reconnect will consider this to be
        SECOND declaration, ie. it will retry ad infinitum
    :type fail_on_first_time_resource_locked: bool
    :param body_receive_mode: how should message.body be received. This
        has a performance impact
    :type body_receive_mode: a property of BodyReceiveMode
    """
    __slots__ = ('queue', 'no_ack', 'on_message', 'cancelled', 'receiver',
                 'attache_group', 'channel_close_sent', 'qos', 'qos_update_sent',
                 'future_to_notify', 'future_to_notify_on_dead',
                 'fail_on_first_time_resource_locked', 'cancel_on_failure',
                 'body_receive_mode', 'consumer_tag', 'on_cancel', 'on_broker_cancel',
                 'hb_watch', 'deliver_watch')

    def __init__(self, queue, on_message, no_ack=True, qos=None,
                 cancel_on_failure=False,
                 future_to_notify=None,
                 fail_on_first_time_resource_locked=False,
                 body_receive_mode=BodyReceiveMode.BYTES
                 ):
        """
        Note that if you specify QoS, it is applied before basic.consume is
        sent. This will prevent the broker from hammering you into oblivion
        with a mountain of messages.
        """
        super(Consumer, self).__init__()

        self.queue = queue
        self.no_ack = no_ack

        self.on_message = on_message

        # consumer?
        self.receiver = None  # MessageReceiver instance

        self.attache_group = None  # attache group this belongs to.
        self.channel_close_sent = False  # for avoiding situations where ChannelClose is sent twice
        # if this is not None, then it has an attribute
        # on_cancel_customer(Consumer instance)
        self.qos = _qosify(qos)
        self.qos_update_sent = False  # QoS was not sent to server

        self.future_to_notify = future_to_notify
        self.future_to_notify_on_dead = None  # .cancel

        self.fail_on_first_time_resource_locked = fail_on_first_time_resource_locked
        self.cancel_on_failure = cancel_on_failure
        self.body_receive_mode = body_receive_mode

        self.consumer_tag = None

        self.on_cancel = Callable(
            oneshots=True)  #: public, called on cancel for any reason
        self.on_broker_cancel = Callable(
            oneshots=True)  #: public, called on Customer Cancel Notification
        #  (RabbitMQ)

    def set_qos(self, prefetch_size, prefetch_count):  # type: (int, int) -> None
        """
        Set new QoS for this consumer.

        :param prefetch_size: prefetch in octets
        :param prefetch_count: prefetch in whole messages
        """
        if self.state == ST_ONLINE:
            self.method(BasicQos(prefetch_size or 0, prefetch_count, False))
        self.qos = prefetch_size or 0, prefetch_count

    def cancel(self):  # type: () -> None
        """
        Cancel the customer.

        .ack() or .nack() for messages from this customer will have no effect.

        :return: a Future to tell when it's done. The future will always
                 succeed - sooner, or later.
                 NOTE: Future is OK'd when entire channel is destroyed
        """

        if self.future_to_notify_on_dead is not None:
            # we cancelled it earlier
            return self.future_to_notify_on_dead
        else:
            self.future_to_notify_on_dead = Future()
            self.future_to_notify_on_dead.set_running_or_notify_cancel()

        self.cancelled = True
        self.on_cancel()
        # you'll blow up big next time you try to use this consumer if you
        # can't cancel, but just close
        if self.consumer_tag is not None:
            if not self.channel_close_sent and self.state == ST_ONLINE:
                self.method_and_watch(BasicCancel(self.consumer_tag, False),
                                      [BasicCancelOk],
                                      self.on_close)
        else:
            if not self.channel_close_sent and self.state == ST_ONLINE:
                self.method(ChannelClose(0, b'cancelling', 0, 0))
                self.channel_close_sent = True

        if self.attache_group is not None:
            self.attache_group.on_cancel_customer(self)

        return self.future_to_notify_on_dead

    def on_operational(self, operational):  # type: (bool) -> None
        super(Consumer, self).on_operational(operational)

        if operational:
            self.channel_close_sent = False
            self.receiver = MessageReceiver(self)

            # notify the future
            if self.future_to_notify is not None:
                self.future_to_notify.set_result(None)
                self.future_to_notify = None

        else:
            self.hb_watch.cancel()
            self.deliver_watch.cancel()
            self.receiver.on_gone()
            self.receiver = None

    def on_close(self, payload=None):
        # type: (tp.Optional[coolamqp.framing.base.AMQPMethodPayload]) -> None
        """
        Handle closing the channel. It sounds like an exception...

        This is done in two steps:
        1. self.state <- ST_OFFLINE, on_event(EV_OFFLINE)   upon detecting
           that no more messages will
           be there
        2. self.channel_id <- None, channel is returned to Connection - c
           hannel has been physically torn down

        Note, this can be called multiple times, and eventually with None.

        """
        if self.cancel_on_failure and (not self.cancelled):
            logger.debug(
                'Consumer is cancel_on_failure and failure seen, True->cancelled')
            self.cancelled = True
            self.on_cancel()

        if self.state == ST_ONLINE:
            # The channel has just lost operationality!
            self.on_operational(False)
            self.state = ST_OFFLINE

            # deliver and head/body watch will clean up after themselves

        should_retry = False

        if isinstance(payload, BasicCancel):
            # Consumer Cancel Notification - by RabbitMQ
            # send them back those memoryviews :D

            # on_close is a one_shot watch. We need to re-register it now.
            self.register_on_close_watch()
            if not self.channel_close_sent:
                self.methods([BasicCancelOk(payload.consumer_tag),
                              ChannelClose(0, b'Received basic.cancel', 0, 0)])
                self.channel_close_sent = True
            self.cancelled = True  # wasn't I?
            self.on_cancel()
            self.on_broker_cancel()
            return

        if isinstance(payload, BasicCancelOk):
            # OK, our cancelling went just fine - proceed with teardown
            self.register_on_close_watch()
            if not self.channel_close_sent:
                self.method(ChannelClose(0, b'Received basic.cancel-ok', 0, 0))
                self.channel_close_sent = True
            return

        if isinstance(payload, ChannelClose):
            rc = payload.reply_code
            if rc == RESOURCE_LOCKED:
                # special handling
                # This is because we might be reconnecting, and the broker
                # doesn't know yet that we are dead.
                # it won't release our exclusive channels, and that's why
                # we'll get RESOURCE_LOCKED.

                if self.fail_on_first_time_resource_locked:
                    # still, a RESOURCE_LOCKED on a first declaration ever
                    # suggests something is very wrong
                    self.cancelled = True
                    self.on_cancel()
                else:
                    # Do not notify the user, and retry at will.
                    # Do not zero the future - we will need to later confirm
                    # it, so it doesn't leak.
                    should_retry = True

            if self.future_to_notify:
                self.future_to_notify.set_exception(AMQPError(payload))
                self.future_to_notify = None
                logger.debug('Notifying connection closed with %s', payload)

        # We might not want to throw the connection away.
        should_retry = should_retry and (not self.cancelled)

        old_con = self.connection

        super(Consumer, self).on_close(
            payload)  # this None's self.connection and returns port
        self.fail_on_first_time_resource_locked = False

        if self.future_to_notify_on_dead:  # notify it was cancelled
            logger.info('Consumer successfully cancelled')
            self.future_to_notify_on_dead.set_result(None)
            self.future_to_notify_on_dead = None
        if should_retry:
            if old_con.state == ST_ONLINE:
                logger.info('Retrying with %s', self.queue.name)
                self.attach(old_con)

    def on_delivery(self, sth):
        """
        Callback for delivery-related shit

        :param sth: AMQPMethodFrame WITH basic-deliver, AMQPHeaderFrame or
            AMQPBodyFrame
        """

        if self.receiver is None:
            # dead, cancelled, whatever
            return

        if isinstance(sth, BasicDeliver):
            self.receiver.on_basic_deliver(sth)
        elif isinstance(sth, AMQPBodyFrame):
            self.receiver.on_body(sth.data)
        elif isinstance(sth, AMQPHeaderFrame):
            self.receiver.on_head(sth)

            # No point in listening for more stuff, that's all the watches
            # even listen for

    def on_setup(self, payload):  # type: (coolamqp.framing.base.AMQPMethodPayload) -> None
        """Called with different kinds of frames - during setup"""

        if isinstance(payload, ChannelOpenOk):
            # Do we need to declare the exchange?

            if self.queue.exchange is not None:
                self.connection.method_and_watch(
                    self.channel_id,
                    ExchangeDeclare(self.queue.exchange.name.encode('utf8'),
                                    self.queue.exchange.type,
                                    False,
                                    self.queue.exchange.durable,
                                    self.queue.exchange.auto_delete,
                                    False,
                                    False,
                                    []),
                    ExchangeDeclareOk,
                    self.on_setup
                )
            else:
                self.on_setup(ExchangeDeclareOk())

        elif isinstance(payload, ExchangeDeclareOk):
            # Declare the queue

            name = b'' if self.queue.anonymous else self.queue.name

            self.connection.method_and_watch(
                self.channel_id,
                QueueDeclare(
                    name,
                    False,
                    self.queue.durable,
                    self.queue.exclusive,
                    self.queue.auto_delete,
                    False,
                    []
                ),
                QueueDeclareOk,
                self.on_setup
            )

        elif isinstance(payload, QueueDeclareOk):
            # did we need an anonymous name?
            if self.queue.anonymous:
                self.queue.name = payload.queue.tobytes()

            # We need any form of binding.
            if self.queue.exchange is not None:
                self.method_and_watch(
                    QueueBind(
                        self.queue.name,
                        self.queue.exchange.name.encode('utf8'),
                        b'', False, []),
                    QueueBindOk,
                    self.on_setup
                )
            else:
                # default exchange, pretend it was bind ok
                self.on_setup(QueueBindOk())
        elif isinstance(payload, QueueBindOk):
            if self.qos is not None:
                self.method_and_watch(
                    BasicQos(self.qos[0], self.qos[1], False),
                    BasicQosOk,
                    self.on_setup
                )
            else:
                self.on_setup(BasicQosOk())  # pretend QoS went ok
        elif isinstance(payload, BasicQosOk):
            self.consumer_tag = uuid.uuid4().hex.encode(
                'utf8')  # str in py2, unicode in py3
            self.method_and_watch(
                BasicConsume(self.queue.name, self.consumer_tag,
                             False, self.no_ack, self.queue.exclusive, False,
                             []),
                BasicConsumeOk,
                self.on_setup
            )
        elif isinstance(payload, BasicConsumeOk):
            # AWWW RIGHT~!!! We're good.
            self.on_operational(True)
            consumer_tag = self.consumer_tag

            # Register watches for receiving shit
            # this is multi-shot by default
            self.hb_watch = HeaderOrBodyWatch(self.channel_id,
                                              self.on_delivery)
            self.connection.watch(self.hb_watch)

            # multi-shot watches need manual cleanup!
            self.deliver_watch = MethodWatch(self.channel_id, BasicDeliver,
                                             self.on_delivery)
            self.deliver_watch.oneshot = False
            self.connection.watch(self.deliver_watch)

            self.state = ST_ONLINE

            if self.cancelled:
                self.method(ChannelClose(0, b'Received basic.cancel-ok', 0, 0))
                self.channel_close_sent = True
                self.state = ST_OFFLINE
                return

            # resend QoS, in case of sth
            if self.qos is not None:
                self.set_qos(self.qos[0], self.qos[1])


def _qosify(qos):
    if qos is not None:
        if isinstance(qos, int):
            qos = 0, qos
        elif qos[0] is None:
            qos = 0, qos[1]  # prefetch_size=0=undefined
    return qos


class MessageReceiver(object):
    """This is an object that is used to received messages.

    It maintains all the state, and is used to ack/nack messages as well.

    This object is TORN DOWN when a consumer goes offline,
    and is recreated when it goes online.

    This is called by consumer upon receiving different parts of the message,
    and may opt to kill the connection on bad framing with
    self.consumer.connection.send(None)
    """
    __slots__ = ('consumer', 'state', 'bdeliver', 'header', 'body', 'data_to_go',
                 'message_size', 'offset', 'acks_pending', 'recv_mode')

    def __init__(self, consumer):  # type: (Consumer) -> None
        self.consumer = consumer
        self.state = 0  # 0 - waiting for Basic-Deliver
        # 1 - waiting for Header
        # 2 - waiting for Body [all]
        # 3 - gone!

        self.bdeliver = None  # payload of Basic-Deliver
        self.header = None  # AMQPHeaderFrame
        if consumer.body_receive_mode == BodyReceiveMode.MEMORYVIEW:
            self.body = None  # None is an important sign - first piece of
            # message
        else:
            self.body = []  # list of payloads
        self.data_to_go = None  # set on receiving header, how much bytes we
        # need yet
        self.message_size = None  # in bytes, of currently received message
        self.offset = 0  # used only in MEMORYVIEW mode - pointer to self.body
        #  (which would be a buffer)

        self.acks_pending = set()  # list of things to ack/reject

        self.recv_mode = consumer.body_receive_mode
        # if BYTES, pieces (as mvs) are received into .body and b''.join()ed
        #     at the end
        # if MEMORYVIEW:
        #                   upon first piece, if it's a single-frame message,
        #                        it's returned at once
        #                   if multiframe, self.body is made into a buffer
        #                        and further are received into it
        # if LIST_OF_MEMORYVIEW, pieces (as mvs) are stored into .body, and
        #     that's returned

    def on_gone(self):
        """Called by Consumer to inform upon discarding this receiver"""
        self.state = 3

    def confirm(self, delivery_tag, success):  # type: (int, tp.Callable[[], None]) -> None
        """
        This crafts a constructor for confirming messages.

        This should return a callable/0, whose calling will ACK or REJECT the
        message.
        Calling it multiple times should have no ill effect.

        If this receiver is long gone,

        :param delivery_tag: delivery_tag to ack
        :param success: True if ACK, False if REJECT
        :return: callable/0
        """

        def callable():
            if self.state == 3:
                return  # Gone!

            if self.consumer.cancelled:
                return  # cancelled!

            if delivery_tag not in self.acks_pending:
                return  # already confirmed/rejected

            if success:
                self.consumer.method(BasicAck(delivery_tag, False))
            else:
                self.consumer.method(BasicReject(delivery_tag, True))

        return callable

    def on_head(self, frame):
        assert self.state == 1
        self.header = frame
        self.message_size = self.data_to_go = frame.body_size
        self.state = 2

        if self.header.body_size == 0:
            # An empty message is no common guest. It won't have a BODY field
            #  though...
            self.on_body(EMPTY_MEMORYVIEW)  # trigger it manually

    def on_basic_deliver(self, payload):
        assert self.state == 0
        self.bdeliver = payload
        self.state = 1

    def on_body(self, payload):
        """:type payload: buffer"""
        assert self.state == 2
        self.data_to_go -= len(payload)

        if self.recv_mode == BodyReceiveMode.MEMORYVIEW:

            if self.body is not None:
                # continuing a multipart
                self.body[self.offset:self.offset + len(payload)] = payload
                self.offset += len(payload)
            else:
                # new one
                if self.data_to_go == 0:  # special case - single frame message
                    self.body = payload
                else:
                    self.body = memoryview(bytearray(self.message_size))
                    self.body[0:len(payload)] = payload
                    self.offset = len(payload)

        else:  # BYTES and LIST_OF_MEMORYVIEW
            self.body.append(payload)

        assert self.data_to_go >= 0

        if self.data_to_go == 0:
            ack_expected = not self.consumer.no_ack

            # Message A-OK!

            if ack_expected:
                self.acks_pending.add(self.bdeliver.delivery_tag)

            from coolamqp.objects import ReceivedMessage

            # Does body need preprocessing?
            body = self.body
            if self.recv_mode == BodyReceiveMode.BYTES:
                if len(self.body) == 1:
                    # common case :)
                    body = self.body[0].tobytes()
                else:
                    # since b''.join() with list comprehension and .tobytes()
                    #  would create
                    # an extra copy of string
                    bio = io.BytesIO()
                    for mv in body:
                        bio.write(mv)

                    body = bio.getvalue()
            # if MEMORYVIEW, then it's already ok

            rm = ReceivedMessage(
                body,
                self.bdeliver.exchange,
                self.bdeliver.routing_key,
                self.header.properties,
                self.bdeliver.delivery_tag,
                None if self.consumer.no_ack else self.confirm(
                    self.bdeliver.delivery_tag, True),
                None if self.consumer.no_ack else self.confirm(
                    self.bdeliver.delivery_tag, False),
            )

            self.consumer.on_message(rm)

            self.state = 0

            # at this point it's safe to clear the body
            if self.recv_mode == BodyReceiveMode.MEMORYVIEW:
                self.body = None
            else:
                self.body = []



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import typing as tp

import monotonic

from coolamqp.framing.frames import AMQPHeartbeatFrame
from coolamqp.uplink.connection.watches import AnyWatch


class Heartbeater(object):
    """
    An object that handles heartbeats
    """

    def __init__(self, connection,  # type: coolamqp.uplink.connection.Connection
                 heartbeat_interval=0  # type: tp.Union[int, float]
                 ):
        self.connection = connection
        self.heartbeat_interval = heartbeat_interval

        self.last_heartbeat_on = monotonic.monotonic()  # last heartbeat from server

        self.connection.watchdog(self.heartbeat_interval, self.on_timer)
        self.connection.watch(AnyWatch(self.on_heartbeat))

    def on_heartbeat(self, frame):
        self.last_heartbeat_on = monotonic.monotonic()

    def on_any_frame(self):
        """
        Hehehe, most AMQP servers are not AMQP-compliant.
        Consider a situation where you just got like a metric shitton of messages,
        and the TCP connection is bustin' filled with those frames.

        Server should still be able to send a heartbeat frame, but it doesn't, because of the queue, and
        BANG, dead.

        I know I'm being picky, but at least I implement this behaviour correctly - see priority argument in send.

        Anyway, we should register an all-watch for this.
        """
        self.last_heartbeat_on = monotonic.monotonic()

    def on_timer(self):
        """Timer says we should send a heartbeat"""
        self.connection.send([AMQPHeartbeatFrame()], priority=True)

        if (
                monotonic.monotonic() - self.last_heartbeat_on) > 2 * self.heartbeat_interval:
            # closing because of heartbeat
            self.connection.send(None)

        self.connection.watchdog(self.heartbeat_interval, self.on_timer)



# coding=UTF-8
"""
Comprehensive management of a framing connection.

Connection is something that can:
 - call something when an AMQPFrame is received
 - send AMQPFrame's

 Pretty much CoolAMQP is about persistent "attaches" that attach to transient connection
 (they die when down) to do stuff, ie. send messages, consume, etc.
"""
from __future__ import absolute_import, division, print_function

from coolamqp.uplink.connection.connection import Connection
from coolamqp.uplink.connection.watches import FailWatch, Watch, \
    HeaderOrBodyWatch, MethodWatch, AnyWatch
from coolamqp.uplink.connection.states import ST_OFFLINE, ST_CONNECTING, \
    ST_ONLINE



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import collections
import socket
import logging

logger = logging.getLogger(__name__)


class SocketFailed(IOError):
    """Failure during socket operation. It needs to be discarded."""


class BaseSocket(object):
    """
    Base class for sockets provided to listeners.

    This is based on a standard TCP socket.

    To be instantiated only by Listeners.
    """

    def __init__(self, sock, on_read=lambda data: None,
                 on_time=lambda: None,
                 on_fail=lambda: None):
        """

        :param sock: socketobject
        :param on_read: callable(data) to be called when data is read.
            Listener thread context
            Raises ValueError on socket should be closed
        :param on_time: callable() when time provided by socket expires
        :param on_fail: callable() when socket is dead and to be discarded.
            Listener thread context.
            Socket descriptor will be handled by listener.
            This should not
        """
        assert sock is not None
        self.sock = sock
        self.data_to_send = collections.deque()
        self.priority_queue = collections.deque()  # when a piece of data is finished, this queue is checked first
        self.my_on_read = on_read
        self._on_fail = on_fail
        self.on_time = on_time
        self.is_failed = False

    def on_fail(self):
        self.is_failed = True
        self._on_fail()

    def send(self, data, priority=True):
        """
        Schedule to send some data.

        :param data: data to send, or None to terminate this socket.
            Note that data will be sent atomically, ie. without interruptions.
        :param priority: preempt other datas. Property of sending data atomically will be maintained.
        """
        if self.is_failed: return

        if data is None:
            # THE POPE OF NOPE
            self.priority_queue = collections.deque()
            self.data_to_send = collections.deque([None])
            return

        if priority:
            self.priority_queue.append(data)
        else:
            self.data_to_send.append(data)

    def oneshot(self, seconds_after, callable):
        """
        Set to fire a callable N seconds after
        :param seconds_after: seconds after this
        :param callable: callable/0
        """
        raise Exception('Abstract; listener should override that')

    def noshot(self):
        """
        Clear all time-delayed callables.

        This will make no time-delayed callables delivered if ran in listener thread
        """
        raise Exception('Abstract; listener should override that')

    def on_read(self):
        """Socket is readable, called by Listener"""
        if self.is_failed: return
        try:
            data = self.sock.recv(2048)
        except (IOError, socket.error) as e:
            raise SocketFailed(repr(e))

        if len(data) == 0:
            raise SocketFailed('connection gracefully closed')

        try:
            self.my_on_read(data)
        except ValueError as e:
            raise SocketFailed(repr(e))

    def wants_to_send_data(self):  # type: () -> bool
        return not (len(self.data_to_send) == 0 and len(self.priority_queue) == 0)

    def on_write(self):
        """
        Socket is writable, called by Listener
        :raises SocketFailed: on socket error
        :return: True if I'm done sending shit for now
        """
        if self.is_failed: return False

        while True:
            if len(self.data_to_send) == 0:
                if len(self.priority_queue) == 0:
                    return True
                else:
                    self.data_to_send.appendleft(self.priority_queue.popleft())

            assert len(self.data_to_send) > 0

            if self.data_to_send[0] is None:
                raise SocketFailed()  # We should terminate the connection!

            try:
                sent = self.sock.send(self.data_to_send[0])
            except (IOError, socket.error):
                raise SocketFailed()

            if sent < len(self.data_to_send[0]):
                # Not everything could be sent
                self.data_to_send[0] = self.data_to_send[0][sent:]
                return False
            else:
                # Looks like everything has been sent
                self.data_to_send.popleft()  # mark as sent

                if len(self.priority_queue) > 0:
                    # We can send a priority pack
                    self.data_to_send.appendleft(self.priority_queue.popleft())

    def fileno(self):  # type: () -> int
        """Return descriptor number"""
        return self.sock.fileno()

    def close(self):
        """Close this socket"""
        self.sock.close()



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import collections
import heapq
import logging
import select
import socket

import monotonic
import six

from coolamqp.uplink.listener.socket import SocketFailed, BaseSocket

logger = logging.getLogger(__name__)

RO = select.EPOLLIN | select.EPOLLHUP | select.EPOLLERR
RW = RO | select.EPOLLOUT


class EpollSocket(BaseSocket):
    """
    EpollListener substitutes your BaseSockets with this
    :type sock: socket.socket
    :type on_read: tp.Callable[[bytes], None]
    :type on_fail: tp.Callable[[], None]
    :type listener: coolamqp.uplink.listener.ListenerThread
    """

    def __init__(self, sock, on_read, on_fail, listener):
        BaseSocket.__init__(self, sock, on_read=on_read, on_fail=on_fail)
        self.listener = listener
        self.priority_queue = collections.deque()

    def send(self, data, priority=False):
        """
        This can actually get called not by ListenerThread.
        """
        BaseSocket.send(self, data, priority=priority)
        try:
            self.listener.epoll.modify(self, RW)
        except socket.error:
            # silence. If there are errors, it's gonna get nuked soon.
            pass

    def oneshot(self, seconds_after, callable):
        """
        Set to fire a callable N seconds after
        :param seconds_after: seconds after this
        :param callable: callable/0
        """
        self.listener.oneshot(self, seconds_after, callable)

    def noshot(self):
        """
        Clear all time-delayed callables.

        This will make no time-delayed callables delivered if ran in listener thread
        """
        self.listener.noshot(self)


class EpollListener(object):
    """
    A listener using epoll.
    """

    def __init__(self):
        self.epoll = select.epoll()
        self.fd_to_sock = {}
        self.time_events = []
        self.sockets_to_activate = []

    def wait(self, timeout=1):
        for socket_to_activate in self.sockets_to_activate:
            logger.debug('Activating fd %s', (socket_to_activate.fileno(),))
            self.epoll.register(socket_to_activate.fileno(), RW)
        self.sockets_to_activate = []

        events = self.epoll.poll(timeout=timeout)

        # Timer events
        mono = monotonic.monotonic()
        while len(self.time_events) > 0 and (self.time_events[0][0] < mono):
            ts, fd, callback = heapq.heappop(self.time_events)
            callback()

        for fd, event in events:
            sock = self.fd_to_sock[fd]

            # Errors
            try:
                if event & (select.EPOLLERR | select.EPOLLHUP):
                    logger.debug('Socket %s has failed', fd)
                    raise SocketFailed()

                if event & select.EPOLLIN:
                    sock.on_read()

                if event & select.EPOLLOUT:

                    sock.on_write()
                    # I'm done with sending for now
                    if len(sock.data_to_send) == 0 and len(
                            sock.priority_queue) == 0:
                        self.epoll.modify(sock.fileno(), RO)

            except SocketFailed as e:
                logger.debug('Socket %s has raised %s', fd, e)
                self.epoll.unregister(fd)
                del self.fd_to_sock[fd]
                sock.on_fail()
                self.noshot(sock)
                sock.close()

        # Do any of the sockets want to send data Re-register them
        for socket in self.fd_to_sock.values():
            if socket.wants_to_send_data():
                self.epoll.modify(socket.fileno(), RW)

    def noshot(self, sock):
        """
        Clear all one-shots for a socket
        :param sock: BaseSocket instance
        """
        fd = sock.fileno()
        self.time_events = [q for q in self.time_events if q[1] != fd]

    def shutdown(self):
        """
        Forcibly close all sockets that this manages (calling their on_fail's),
        and close the object.

        This object is unusable after this call.
        """
        self.time_events = []
        for sock in list(six.itervalues(self.fd_to_sock)):
            sock.on_fail()
            sock.close()

        self.fd_to_sock = {}
        self.epoll.close()

    def oneshot(self, sock, delta, callback):
        """
        A socket registers a time callback
        :param sock: BaseSocket instance
        :param delta: "this seconds after now"
        :param callback: callable/0
        """
        if sock.fileno() in self.fd_to_sock:
            heapq.heappush(self.time_events, (monotonic.monotonic() + delta,
                                              sock.fileno(),
                                              callback
                                              ))

    def activate(self, sock):  # type: (coolamqp.uplink.listener.epoll_listener.EpollSocket) -> None
        self.sockets_to_activate.append(sock)

    def register(self, sock, on_read=lambda data: None,
                 on_fail=lambda: None):
        """
        Add a socket to be listened for by the loop.

        :param sock: a socket instance (as returned by socket module)
        :param on_read: callable(data) to be called with received data
        :param on_fail: callable() to be called when socket fails

        :return: a BaseSocket instance to use instead of this socket
        """
        sock = EpollSocket(sock, on_read, on_fail, self)
        self.fd_to_sock[sock.fileno()] = sock

        return sock



# coding=UTF-8
from __future__ import absolute_import, division, print_function

"""
- Compile XML to definitions.py
- Dynamically compile classes from property_flags
"""

from .content_property import compile_particular_content_property_list_class



# coding=UTF-8
"""
queue.declare, exchange.declare and that shit
"""
from __future__ import print_function, absolute_import, division

import collections
import logging
from concurrent.futures import Future

from coolamqp.attaches.channeler import Channeler, ST_ONLINE
from coolamqp.attaches.utils import Synchronized
from coolamqp.exceptions import AMQPError, ConnectionDead
from coolamqp.framing.definitions import ChannelOpenOk, ExchangeDeclare, \
    ExchangeDeclareOk, QueueDeclare, \
    QueueDeclareOk, ChannelClose, QueueDelete, QueueDeleteOk
from coolamqp.objects import Exchange, Queue, Callable

logger = logging.getLogger(__name__)


class Operation(object):
    """
    An abstract operation.

    This class possesses the means to carry itself out and report back status.
    Represents the op currently carried out.

    This will register it's own callback. Please, call on_connection_dead when connection is broken
    to fail futures with ConnectionDead, since this object does not watch for Fails
    """
    __slots__ = ('done', 'fut', 'declarer', 'obj', 'on_done')

    def __init__(self, declarer, obj, fut=None):
        self.done = False
        self.fut = fut
        self.declarer = declarer
        self.obj = obj

        self.on_done = Callable()  # callable/0

    def on_connection_dead(self):
        """To be called by declarer when our link fails"""
        if self.fut is not None:
            self.fut.set_exception(ConnectionDead())
            self.fut = None

    def perform(self):
        """Attempt to perform this op."""
        obj = self.obj
        if isinstance(obj, Exchange):
            self.declarer.method_and_watch(
                ExchangeDeclare(self.obj.name.encode('utf8'), obj.type, False,
                                obj.durable,
                                obj.auto_delete, False, False, []),
                (ExchangeDeclareOk, ChannelClose),
                self._callback)
        elif isinstance(obj, Queue):
            self.declarer.method_and_watch(
                QueueDeclare(obj.name, False, obj.durable, obj.exclusive,
                             obj.auto_delete, False, []),
                (QueueDeclareOk, ChannelClose),
                self._callback)

    def _callback(self, payload):
        assert not self.done
        self.done = True
        if isinstance(payload, ChannelClose):
            if self.fut is not None:
                self.fut.set_exception(AMQPError(payload))
                self.fut = None
            else:
                # something that had no Future failed. Is it in declared?
                if self.obj in self.declarer.declared:
                    self.declarer.declared.remove(
                        self.obj)  # todo access not threadsafe
                    self.declarer.on_discard(self.obj)
        else:
            if self.fut is not None:
                self.fut.set_result(None)
                self.fut = None
        self.declarer.on_operation_done()


class DeleteQueue(Operation):
    def __init__(self, declarer, queue, fut):
        super(DeleteQueue, self).__init__(declarer, queue, fut=fut)

    def perform(self):
        queue = self.obj

        self.declarer.method_and_watch(
            QueueDelete(queue.name, False, False, False),
            (QueueDeleteOk, ChannelClose),
            self._callback)

    def _callback(self, payload):
        assert not self.done
        self.done = True
        if isinstance(payload, ChannelClose):
            self.fut.set_exception(AMQPError(payload))
        else:  # Queue.DeleteOk
            self.fut.set_result(None)
        self.declarer.on_operation_done()


class Declarer(Channeler, Synchronized):
    """
    Doing other things, such as declaring, deleting and other stuff.

    This also maintains a list of declared queues/exchanges, and redeclares them on each reconnect.
    """

    def __init__(self):
        """
        Create a new declarer.
        """
        Channeler.__init__(self)
        Synchronized.__init__(self)

        self.declared = set()  # since Queues and Exchanges are hashable...
        # anonymous queues aren't, but we reject those
        # persistent

        self.left_to_declare = collections.deque()  # since last disconnect. persistent+transient
        # deque of Operation objects

        self.on_discard = Callable()  # callable/1, with discarded elements

        self.in_process = None  # Operation instance that is being progressed right now

    def on_close(self, payload=None):

        # we are interested in ChannelClose during order execution,
        # because that means that operation was illegal, and must
        # be discarded/exceptioned on future

        if payload is None:

            if self.in_process is not None:
                self.in_process.on_connection_dead()
                self.in_process = None

            # connection down, panic mode engaged.
            while len(self.left_to_declare) > 0:
                self.left_to_declare.pop().on_connection_dead()

            # recast current declarations as new operations
            for dec in self.declared:
                self.left_to_declare.append(Operation(self, dec))

            super(Declarer, self).on_close()
            return

        elif isinstance(payload, ChannelClose):
            # Looks like a soft fail - we may try to survive that
            old_con = self.connection
            super(Declarer, self).on_close()

            # But, we are super optimists. If we are not cancelled, and connection is ok,
            # we must reestablish
            if old_con.state == ST_ONLINE and not self.cancelled:
                self.attach(old_con)
        else:
            super(Declarer, self).on_close(payload)

    def on_operation_done(self):
        """
        Called by operation, when it's complete (whether success or fail).
        Not called when operation fails due to DC
        """
        self.in_process = None
        self._do_operations()

    def delete_queue(self, queue):
        """
        Delete a queue.

        Future is returned, so that user knows when it happens. This may fail.
        Returned Future is already running, and so cannot be cancelled.

        If the queue is in declared consumer list, it will not be removed.

        :param queue: Queue instance
        :return: a Future
        """
        fut = Future()
        fut.set_running_or_notify_cancel()

        self.left_to_declare.append(DeleteQueue(self, queue, fut))
        self._do_operations()

        return fut

    def declare(self, obj, persistent=False):
        """
        Schedule to have an object declared.

        Future is returned, so that user knows when it happens.
        Returned Future is already running, and so cannot be cancelled.

        Exchange declarations never fail.
            Of course they do, but you will be told that it succeeded. This is by design,
            and due to how AMQP works.

        Queue declarations CAN fail.

        Note that if re-declaring these fails, they will be silently discarded.
        You can subscribe an on_discard(Exchange | Queue) here.

        :param obj: Exchange or Queue instance
        :param persistent: will be redeclared upon disconnect. To remove, use "undeclare"
        :return: a Future instance
        :raise ValueError: tried to declare anonymous queue
        """
        if isinstance(obj, Queue):
            if obj.anonymous:
                raise ValueError('Cannot declare anonymous queue')

        fut = Future()
        fut.set_running_or_notify_cancel()

        if persistent:
            if obj not in self.declared:
                self.declared.add(obj)  # todo access not threadsafe

        self.left_to_declare.append(Operation(self, obj, fut))
        self._do_operations()

        return fut

    @Synchronized.synchronized
    def _do_operations(self):
        """
        Attempt to execute something.

        To be called when it's possible that something can be done
        """
        if (self.state != ST_ONLINE) or len(self.left_to_declare) == 0 or (
                    self.in_process is not None):
            return

        self.in_process = self.left_to_declare.popleft()
        self.in_process.perform()

    def on_setup(self, payload):
        if isinstance(payload, ChannelOpenOk):
            assert self.in_process is None
            self.state = ST_ONLINE
            self._do_operations()



# coding=UTF-8
from __future__ import absolute_import, division, print_function

from coolamqp.framing.definitions import HARD_ERRORS, RESOURCE_LOCKED

__all__ = ['HARD_ERRORS', 'RESOURCE_LOCKED', 'CoolAMQPError', 'ConnectionDead', 'AMQPError']


class CoolAMQPError(Exception):
    """Base class for CoolAMQP errors"""


class ConnectionDead(CoolAMQPError):
    """
    Operation could be not completed because some other error
    than a legit AMQPError occurred, such as exploding kitten
    """


class AMQPError(CoolAMQPError):
    """
    Base class for errors received from AMQP server
    """

    def is_hard_error(self):  # type: () -> bool
        """Does this error close the connection?"""
        return self.reply_code in HARD_ERRORS

    def __str__(self):  # type: () -> str
        return 'AMQP error %s: %s' % (self.reply_code, self.__get_reply_text())

    def __get_reply_text(self):  # type: () -> str
        if isinstance(self.reply_text, memoryview):
            reply_text = self.reply_text.tobytes().decode('utf8')
        else:
            reply_text = self.reply_text

        if isinstance(reply_text, bytes):
            reply_text = self.reply_text.decode('utf8')

        return reply_text

    def __repr__(self):  # type: () -> str
        return 'AMQPError(' + repr(self.reply_code) + ', ' + repr(
            self.__get_reply_text()) + \
               ', ' + repr(self.class_id) + ', ' + repr(self.method_id) + ')'

    def __init__(self, *args):
        """

        :param args: can be either reply_code, reply_text, class_id, method_id
                     or a ConnectionClose/ChannelClose.
        """
        from coolamqp.framing.definitions import ConnectionClose, ChannelClose

        if isinstance(args[0], (ConnectionClose, ChannelClose)):
            self.reply_code = args[0].reply_code
            self.reply_text = args[0].reply_text
            self.class_id = args[0].class_id
            self.method_id = args[0].method_id
        else:
            assert len(args) == 4
            self.reply_code, self.reply_text, self.class_id, self.method_id = args



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import threading
import typing as tp

from coolamqp.objects import Callable
from coolamqp.uplink.listener.epoll_listener import EpollListener


class ListenerThread(threading.Thread):
    """
    A thread that does the listening.

    It automatically picks the best listener for given platform.
    """

    def __init__(self, name=None):  # type: (tp.Optional[str]) -> None
        threading.Thread.__init__(self, name='coolamqp/ListenerThread')
        self.daemon = True
        self.name = name or 'CoolAMQP'
        self.terminating = False
        self._call_next_io_event = Callable(oneshots=True)

    def call_next_io_event(self, callable):
        """
        Call callable after current I/O event is fully processed

        sometimes many callables are called in response to single
        I/O (eg. teardown, startup). This guarantees a call after
        all these are done.
        :param callable: callable/0
        """
        self._call_next_io_event()

    def terminate(self):
        self.terminating = True

    def init(self):
        """Called before start. It is not safe to fork after this"""
        self.listener = EpollListener()

    def activate(self, sock):
        self.listener.activate(sock)

    def run(self):
        try:
            import prctl
        except ImportError:
            pass
        else:
            prctl.set_name(self.name+' - AMQP listener thread')

        while not self.terminating:
            self.listener.wait(timeout=1)

        self.listener.shutdown()

    def register(self, sock, on_read=lambda data: None, on_fail=lambda: None):
        """
        Add a socket to be listened for by the loop.

        :param sock: a socket instance (as returned by socket module)
        :param on_read: callable(data) to be called with received data
        :param on_fail: callable() to be called when socket fails

        :return: a BaseSocket instance to use instead of this socket
        """
        return self.listener.register(sock, on_read, on_fail)



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import typing as tp

AMQP_HELLO_HEADER = b'AMQP\x00\x00\x09\x01'


# name => (length|None, struct ID|None, reserved-field-value : for struct if structable, bytes else, length of default)
BASIC_TYPES = {u'bit': (None, None, "0", None),  # special case
               u'octet': (1, 'B', "b'\\x00'", 1),
               u'short': (2, 'H', "b'\\x00\\x00'", 2),
               u'long': (4, 'I', "b'\\x00\\x00\\x00\\x00'", 4),
               u'longlong': (
                   8, 'Q', "b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00'", 8),
               u'timestamp': (
                   8, 'Q', "b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00'", 8),
               u'table': (None, None, "b'\\x00\\x00\\x00\\x00'", 4),
               # special case
               u'longstr': (None, None, "b'\\x00\\x00\\x00\\x00'", 4),
               # special case
               u'shortstr': (None, None, "b'\\x00'", 1),  # special case
               }

DYNAMIC_BASIC_TYPES = (u'table', u'longstr', u'shortstr')


class AMQPFrame(object):  # base class for framing
    FRAME_TYPE = None  # override me!

    def __init__(self, channel):  # type: (int) -> None
        self.channel = channel

    def write_to(self, buf):  # type: (io.BinaryIO) -> None
        """
        Write a complete frame to buffer

        This writes type and channel ID.
        """
        # DO NOT UNCOMMENT buf.write(struct.pack('!BH', self.FRAME_TYPE, self.channel))
        raise NotImplementedError(
            'Please write the frame type and channel in child classes, its faster that way ')

    @staticmethod
    def unserialize(channel, payload_as_buffer):
        """
        Unserialize from a buffer.
        Buffer starts at frame's own payload - type, channel and size was already obtained.
        Payload does not contain FRAME_EMD.
        AMQPHeartbeatFrame does not have to implement this.
        """
        raise NotImplementedError('Override me')

    def get_size(self):  # type: () -> int
        """
        Return size of this frame, in bytes, from frame type to frame_end
        :return: int
        """
        raise NotImplementedError('Override me')

    def __str__(self):  # type: () -> str
        return 'AMQPFrame(%s)' % (self.channel,)


class AMQPPayload(object):
    """Payload is something that can write itself to bytes,
    or at least provide a buffer to do it."""
    __slots__ = ()

    def write_to(self, buf):  # type: (buffer) -> None
        """
        Emit itself into a buffer, from length to FRAME_END

        :param buf: buffer to write to (will be written using .write)
        """

    def get_size(self):  # type: () -> int
        """
        Return size of this payload
        :return: int
        """
        raise NotImplementedError()


class AMQPClass(object):
    """An AMQP class"""

    __slots__ = ()


class AMQPContentPropertyList(object):
    """
    A class is intmately bound with content and content properties.

    WARNING: BE PREPARED that if you receive a content from the network,
    string values will be memoryviews. Use .tobytes() to correct that.
    If YOU create a property list, they will be bytes all right.
    """
    PROPERTIES = []

    # todo they are immutable, so they could just serialize themselves...

    def __str__(self):  # type: () -> str
        return '<AMQPContentPropertyList>'

    def get(self, property_name, default=None):
        # type: (str, str) -> tp.Union[memoryview, bytes]
        """
        Return a particular property, or default if not defined

        :param property_name: property name, unicode
        :param default: default value
        :return: memoryview or bytes
        """
        return getattr(self, property_name, default)

    @staticmethod
    def zero_property_flags(property_flags):
        """
        Given a binary property_flags, set all bit properties to 0.

        This leaves us with a canonical representation, that can be used
        in obtaining a particular property list

        :param property_flags: binary
        :return: binary
        """
        # this is a default implementation.
        # compiler should emit it's own when the content property list has a
        # possible bit field
        return property_flags

    def write_to(self, buf):
        """Serialize itself (flags + values) to a buffer"""
        raise Exception(u'This is an abstract method')

    @staticmethod
    def from_buffer(self, buf, start_offset):  # type: (buffer, int) -> AMQPContentPropertyList
        """
        Return an instance of self, loaded from a buffer.

        This does not have to return length, because it is always passed exactly enough of a buffer.

        Buffer HAS TO start at property_flags
        """
        raise Exception(u'This is an abstract method')

    def get_size(self):  # type: () -> int
        """
        How long is property_flags + property_values

        :return: int
        """
        raise Exception(u'This is an abstract method')


class AMQPMethodPayload(AMQPPayload):
    __slots__ = ()

    RESPONSE_TO = None
    REPLY_WITH = []
    FIELDS = []

    def get_size(self):  # type: () -> int
        """
        Calculate the size of this frame.

        Needs to be overloaded, unless you're a class with IS_CONTENT_STATIC

        :return: int, size of argument section
        :raises RuntimeError: this class isn't IS_CONTENT_STATIC and this method was called directly.
            In this case, you should have rather subclassed it.
        """
        if self.IS_CONTENT_STATIC:
            return len(
                self.STATIC_CONTENT) - 4 - 4 - 1  # minus length, class, method, frame_end

        raise RuntimeError('Should never be executed!')

    def write_arguments(self, buf):
        """
        Write the argument portion of this frame into target buffer.

        :param buf: buffer to write to
        :type buf: tp.BinaryIO
        :raise ValueError: some field here is invalid!
        """
        raise NotImplementedError()

    @classmethod
    def from_buffer(cls, buf, offset):  # type: (buffer, int) -> AMQPMethodPayload
        """
        Construct this frame from a buffer

        :param buf: a buffer to construct the frame from
        :type buf: buffer or memoryview
        :param offset: offset the argument portion begins at
        :type offset: int
        :return: an instance of this class
        :raise ValueError: invalid data
        """
        raise NotImplementedError('')
